from django.db.backends.base.creation import BaseDatabaseCreation


class DatabaseCreation(BaseDatabaseCreation):

    def create_test_db(self, *args, **kwargs):
        import os
        db_name = super().create_test_db()

        here = os.path.realpath(os.path.join(os.getcwd(), os.path.dirname(__file__)))
        with self.connection.cursor() as cursor:
            for f in ['init.sql', 'clr.sql']:
                with open(f'{here}/sql/{f}', 'r') as file:
                    sql = file.read()
                    for s in sql.split('\nGO\n'):
                        cursor.execute(s)
        return db_name

    def sql_table_creation_suffix(self):
        """ a lot of tests expect case sensitivity """
        return 'COLLATE Latin1_General_100_CS_AS_SC '



import binascii
import datetime
import copy

from django.db.backends.utils import strip_quotes
from django.apps.registry import Apps
from django.db.backends.base.schema import BaseDatabaseSchemaEditor, _related_non_m2m_objects
from django.utils.encoding import force_str
from django.db.backends.ddl_references import Statement

from tds_django.sql.queries import Misc


class DatabaseSchemaEditor(BaseDatabaseSchemaEditor):
    sql_create_column = 'ALTER TABLE %(table)s ADD %(column)s %(definition)s'
    sql_alter_column_default = 'ADD DEFAULT %(default)s FOR %(column)s'
    sql_delete_column = 'ALTER TABLE %(table)s DROP COLUMN %(column)s'
    sql_alter_column_no_default = 'DROP CONSTRAINT IF EXISTS %(column)s'
    sql_alter_column_type = 'ALTER COLUMN %(column)s %(type)s'
    sql_alter_column_collate = 'ALTER COLUMN %(column)s %(type)s %(collation)s'
    sql_delete_index = 'DROP INDEX %(name)s ON %(table)s'
    sql_delete_table = Misc.delete_table
    sql_rename_column = "EXEC sp_rename '%(table)s.%(old_column)s', %(new_column)s, 'COLUMN'"
    sql_rename_table = 'EXEC sp_rename %(old_table)s, %(new_table)s'

    sql_alter_column_not_null = 'ALTER COLUMN %(column)s %(type)s NOT NULL'
    sql_alter_column_null = 'ALTER COLUMN %(column)s %(type)s NULL'

    sql_drop_constraint = 'ALTER TABLE %(table)s DROP CONSTRAINT IF EXISTS %(name)s'

    _auto_field_types = {'AutoField', 'BigAutoField', 'SmallAutoField'}

    def prepare_default(self, value):
        return self.quote_value(value)

    @staticmethod
    def quote_value(value):
        """
        Returns a quoted version of the value so it's safe to use in an SQL
        string. This is not safe against injection from user code; it is
        intended only for use in making SQL scripts or preparing default values
        for particularly tricky backends (defaults are not user-defined, though,
        so this is safe).
        """
        if isinstance(value, (datetime.datetime, datetime.date, datetime.time)):
            return "'%s'" % value
        elif isinstance(value, str):
            return "N'%s'" % value.replace("'", "''")
        elif isinstance(value, (bytes, bytearray, memoryview)):
            return "0x%s" % force_str(binascii.hexlify(value))
        elif isinstance(value, bool):
            return '1' if value else '0'
        else:
            return str(value)

    def _alter_column_default_sql(self, model, old_field, new_field, drop=False):
        if drop:
            with self.connection.cursor() as cursor:
                defaults = self.connection.introspection.get_table_defaults(cursor, model._meta.db_table)
                constraint_name = defaults.get(new_field.column, ['DEF_DOES_NOT_EXISTS'])[0]
            return (self.sql_alter_column_no_default % {
                'column': constraint_name,
            }, [])
        return super()._alter_column_default_sql(model, old_field, new_field)

    def _is_identity_column(self, table_name, column_name):
        with self.connection.cursor() as cursor:
            sequences = self.connection.introspection.get_sequences(cursor, table_name)
        return column_name in [s['column'] for s in sequences]

    def _drop_pk(self, table_name, column_name):
        with self.connection.cursor() as cursor:
            constraints = self.connection.introspection.get_constraints(cursor, table_name)
            # drop primary key
            for name, props in constraints.items():
                if props['primary_key'] and column_name in props['columns']:
                    self.execute(self.sql_drop_constraint % {
                        'table': self.quote_name(table_name),
                        'name': self.quote_name(name)
                    })
                    # django 3.2 only supports one pk
                    break

    def _set_field_new_type_null_status(self, field, new_type):
        """ same mysql """
        if field.null:
            new_type += ' NULL'
        else:
            new_type += ' NOT NULL'
        return new_type

    def _alter_field(self, model, old_field, new_field, old_type, new_type,
                     old_db_params, new_db_params, strict=False):
        old_internal = old_field.get_internal_type()
        new_internal = new_field.get_internal_type()
        if (old_internal in self._auto_field_types) != (new_internal in self._auto_field_types):
            # need to recreate either column or table
            # choose remaking table like sqlite yolo
            # delete FKs and their index
            for _old_rel, new_rel in _related_non_m2m_objects(old_field, new_field):
                self._remove_constraints(new_rel.related_model, new_rel.field.column)

            self._remake_table(model, alter_field=(old_field, new_field))

            # Rebuild tables with FKs pointing to this field.
            if new_field.unique and old_type != new_type:
                related_models = set()
                opts = new_field.model._meta
                for remote_field in opts.related_objects:
                    # Ignore self-relationship since the table was already rebuilt.
                    if remote_field.related_model == model:
                        continue
                    if not remote_field.many_to_many:
                        if remote_field.field_name == new_field.name:
                            related_models.add(remote_field.related_model)
                    elif new_field.primary_key and remote_field.through._meta.auto_created:
                        related_models.add(remote_field.through)
                if new_field.primary_key:
                    for many_to_many in opts.many_to_many:
                        # Ignore self-relationship since the table was already rebuilt.
                        if many_to_many.related_model == model:
                            continue
                        if many_to_many.remote_field.through._meta.auto_created:
                            related_models.add(many_to_many.remote_field.through)
                for related_model in related_models:
                    self._remake_table(related_model)
        else:
            # null does not allow "post actions"
            post_actions = []
            if old_field.null and not new_field.null:
                # drop and recreate indices
                post_actions = self._remove_constraints(model, old_field.column, new_field, index=True, pk=False,
                                                        fk=False, check=False, unique=False)
            super()._alter_field(model, old_field, new_field, old_type, new_type, old_db_params, new_db_params, strict)
            if self._field_becomes_null_unique(old_field, new_field):
                column = new_field.column
                condition = '%s IS NOT NULL' % self.quote_name(column)
                sql = self._create_unique_sql(model, [new_field], condition=condition)
                post_actions.append(sql)

            for sql in post_actions:
                self.execute(sql)

    def remove_field(self, model, field):
        """
        Remove a field from a model. Usually involves deleting a column,
        but for M2Ms may involve deleting a table.
        """
        # Special-case implicit M2M tables
        if field.many_to_many and field.remote_field.through._meta.auto_created:
            return self.delete_model(field.remote_field.through)
        # It might not actually have a column behind it
        if field.db_parameters(connection=self.connection)['type'] is None:
            return
        # PATCH START
        self._remove_constraints(model, field.column)
        # PATCH END
        # Delete the column
        sql = self.sql_delete_column % {
            'table': self.quote_name(model._meta.db_table),
            'column': self.quote_name(field.column),
        }
        self.execute(sql)
        # Reset connection if required
        if self.connection.features.connection_persists_old_columns:
            self.connection.close()
        # Remove all deferred statements referencing the deleted column.
        for sql in list(self.deferred_sql):
            if isinstance(sql, Statement) and sql.references_column(model._meta.db_table, field.column):
                self.deferred_sql.remove(sql)

    def _remake_table(self, model, create_field=None, delete_field=None, alter_field=None):
        """
        Similar to sqlite
        """
        # Self-referential fields must be recreated rather than copied from
        # the old model to ensure their remote_field.field_name doesn't refer
        # to an altered field.
        def is_self_referential(f):
            return f.is_relation and f.remote_field.model is model

        # Work out the new fields dict / mapping
        body = {
            f.name: f.clone() if is_self_referential(f) else f
            for f in model._meta.local_concrete_fields
        }
        # Since mapping might mix column names and default values,
        # its values must be already quoted.
        mapping = {f.column: self.quote_name(f.column) for f in model._meta.local_concrete_fields}
        # This maps field names (not columns) for things like unique_together
        rename_mapping = {}
        # If any of the new or altered fields is introducing a new PK,
        # remove the old one
        restore_pk_field = None
        if getattr(create_field, 'primary_key', False) or (
                alter_field and getattr(alter_field[1], 'primary_key', False)):
            for name, field in list(body.items()):
                if field.primary_key:
                    field.primary_key = False
                    restore_pk_field = field
                    if field.auto_created:
                        del body[name]
                        del mapping[field.column]
        # Add in any created fields
        if create_field:
            body[create_field.name] = create_field
            # Choose a default and insert it into the copy map
            if not create_field.many_to_many and create_field.concrete:
                mapping[create_field.column] = self.quote_value(
                    self.effective_default(create_field)
                )
        # Add in any altered fields
        if alter_field:
            old_field, new_field = alter_field
            body.pop(old_field.name, None)
            mapping.pop(old_field.column, None)
            body[new_field.name] = new_field
            if old_field.null and not new_field.null:
                case_sql = 'coalesce(%(col)s, %(default)s)' % {
                    'col': self.quote_name(old_field.column),
                    'default': self.quote_value(self.effective_default(new_field))
                }
                mapping[new_field.column] = case_sql
            else:
                mapping[new_field.column] = self.quote_name(old_field.column)
            rename_mapping[old_field.name] = new_field.name
        # Remove any deleted fields
        if delete_field:
            del body[delete_field.name]
            del mapping[delete_field.column]
            # Remove any implicit M2M tables
            if delete_field.many_to_many and delete_field.remote_field.through._meta.auto_created:
                return self.delete_model(delete_field.remote_field.through)
        # Work inside a new app registry
        apps = Apps()

        # Work out the new value of unique_together, taking renames into
        # account
        unique_together = [
            [rename_mapping.get(n, n) for n in unique]
            for unique in model._meta.unique_together
        ]

        # Work out the new value for index_together, taking renames into
        # account
        index_together = [
            [rename_mapping.get(n, n) for n in index]
            for index in model._meta.index_together
        ]

        indexes = model._meta.indexes
        if delete_field:
            indexes = [
                index for index in indexes
                if delete_field.name not in index.fields
            ]

        constraints = list(model._meta.constraints)

        # Provide isolated instances of the fields to the new model body so
        # that the existing model's internals aren't interfered with when
        # the dummy model is constructed.
        body_copy = copy.deepcopy(body)

        # Construct a new model with the new fields to allow self referential
        # primary key to resolve to. This model won't ever be materialized as a
        # table and solely exists for foreign key reference resolution purposes.
        # This wouldn't be required if the schema editor was operating on model
        # states instead of rendered models.
        meta_contents = {
            'app_label': model._meta.app_label,
            'db_table': model._meta.db_table,
            'unique_together': unique_together,
            'index_together': index_together,
            'indexes': indexes,
            'constraints': constraints,
            'apps': apps,
        }
        meta = type('Meta', (), meta_contents)
        body_copy['Meta'] = meta
        body_copy['__module__'] = model.__module__
        type(model._meta.object_name, model.__bases__, body_copy)

        # Construct a model with a renamed table name.
        body_copy = copy.deepcopy(body)
        meta_contents = {
            'app_label': model._meta.app_label,
            'db_table': 'new__%s' % strip_quotes(model._meta.db_table),
            'unique_together': unique_together,
            'index_together': index_together,
            'indexes': indexes,
            'constraints': constraints,
            'apps': apps,
        }
        meta = type('Meta', (), meta_contents)
        body_copy['Meta'] = meta
        body_copy['__module__'] = model.__module__
        new_model = type('New%s' % model._meta.object_name, model.__bases__, body_copy)

        # Create a new table with the updated schema.
        self.create_model(new_model)

        # Copy data from the old table into the new table
        sql = 'INSERT INTO %(table)s (%(source)s) SELECT %(dest)s FROM %(old_table)s' % {
            'table': self.quote_name(new_model._meta.db_table),
            'source': ', '.join(self.quote_name(x) for x in mapping),
            'dest': ', '.join(mapping.values()),
            'old_table': self.quote_name(model._meta.db_table),
        }
        if new_model._meta.auto_field:
            sql = 'SET IDENTITY_INSERT %(table)s ON;%(sql)s; SET IDENTITY_INSERT %(table)s OFF' % {
                'sql': sql,
                'table': self.quote_name(new_model._meta.db_table),
            }
        self.execute(sql)

        # Delete the old table to make way for the new
        self.delete_model(model, handle_autom2m=False)

        # Rename the new table to take way for the old
        self.alter_db_table(new_model, new_model._meta.db_table, model._meta.db_table)  # , disable_constraints=False,

        # Run deferred SQL on correct table
        for sql in self.deferred_sql:
            self.execute(sql)
        self.deferred_sql = []
        # Fix any PK-removed field
        if restore_pk_field:
            restore_pk_field.primary_key = True

    def delete_model(self, model, handle_autom2m=True):
        if handle_autom2m:
            super().delete_model(model)
        else:
            # Delete the table (and only that)
            self.execute(self.sql_delete_table % {
                'table': self.quote_name(model._meta.db_table),
            })
            # Remove all deferred statements referencing the deleted table.
            for sql in list(self.deferred_sql):
                if isinstance(sql, Statement) and sql.references_table(model._meta.db_table):
                    self.deferred_sql.remove(sql)

    def _remove_constraints(self, model, column_name, new_field=None, index=True, pk=True, unique=True, fk=True,
                            check=True):
        """ remove constraints and return constraint creation sql """
        table_name = model._meta.db_table
        reverse = []
        new_db_params = new_field.db_parameters(connection=self.connection) if new_field else None
        with self.connection.cursor() as cursor:
            constraints = self.connection.introspection.get_constraints(cursor, table_name)
        for name, info in constraints.items():
            if column_name in info['columns']:
                sql = None
                if info['foreign_key'] and fk:
                    sql = self.sql_drop_constraint
                elif info['primary_key'] and pk:
                    sql = self.sql_delete_pk
                    if new_field and new_field.primary_key:
                        reverse.append(self.sql_create_pk % {
                            'table': self.quote_name(table_name), 'name': self.quote_name(name),
                            'columns': self.quote_name(new_field.column)
                        })
                elif info['check'] and check:
                    sql = self.sql_delete_check
                    if new_field and new_db_params['check']:
                        clause = info['clause']
                        if column_name != new_field.column:
                            clause = clause.replace(self.quote_name(column_name), self.quote_name(new_field.column))
                        reverse.append(self.sql_create_check % {
                            'table': self.quote_name(table_name), 'name': self.quote_name(name),
                            'check': clause,
                        })
                elif info['unique'] and unique:
                    sql = self.sql_delete_unique
                    if new_field:
                        columns = [model._meta.get_field(c) if c != column_name else new_field for c in info['columns']]
                        reverse.append(self._create_unique_sql(model, columns, name))
                elif info['index'] and index:
                    sql = self.sql_delete_index
                    if new_field and not (len(info['columns']) == 1 and info['unique'] and not new_field.unique):
                        columns = [c if c != column_name else new_field.column for c in info['columns']]
                        # from null unique to unique
                        if not new_field.null and len(columns) == 1 and info['unique'] and info['has_filter'] and \
                                ('(%s IS NOT NULL)' % self.quote_name(column_name)) == info['filter_definition']:
                            condition = ''
                        else:
                            condition = f" WHERE {info['filter_definition']}" if info['has_filter'] else ''
                        reverse.append('CREATE %s INDEX %s ON %s (%s)%s' % (
                            'UNIQUE' if info['unique'] else '',
                            self.quote_name(name),
                            self.quote_name(table_name),
                            ', '.join(self.quote_name(c) for c in columns),
                            condition,
                        ))
                if sql:
                    self.execute(sql % {
                        'table': self.quote_name(table_name),
                        'name': self.quote_name(name),
                    })
        return reverse if new_field else None

    def _alter_column_type_sql(self, model, old_field, new_field, new_type):
        fragment, more = super()._alter_column_type_sql(model, old_field, new_field, new_type)
        if not new_field.null:
            (sql, params) = fragment
            fragment = (sql + ' NOT NULL', params)
        # at this point the field would have already be renamed
        todo = self._remove_constraints(model, new_field.column, new_field)
        more += [(sql, []) for sql in todo]
        return fragment, more

    def _collate_sql(self, collation):
        return ' COLLATE ' + collation

    def _field_should_be_indexed(self, model, field):
        """ we prevent the nullable unique constraint at table creation so we need to do it now """
        create_index = super()._field_should_be_indexed(model, field) or (field.unique and field.null)
        if create_index:
            db_type = field.db_type(self.connection)
            if db_type is not None and db_type.lower() in 'nvarchar(max)':
                return False
        return create_index

    def column_sql(self, model, field, include_default=False):
        """ we need to delay nullable unique """
        if field.unique and field.null:
            field._unique = False
            sql, params = super().column_sql(model, field, include_default)
            field._unique = True
            return sql, params
        return super().column_sql(model, field, include_default)

    def _create_index_sql(self, model, *, fields=None, sql=None, suffix='', **kwargs):
        """ for nullable unique constraint """
        if not sql and len(fields) == 1 and fields[0].unique and fields[0].null:
            column = fields[0].column
            condition = '%s IS NOT NULL' % self.quote_name(column)
            return self._create_unique_sql(model, fields, condition=condition)
        return super()._create_index_sql(model, fields=fields, sql=sql, suffix=suffix, **kwargs)

    def _unique_should_be_added(self, old_field, new_field):
        if self._field_becomes_null_unique(old_field, new_field):
            return False
        return super()._unique_should_be_added(old_field, new_field)

    def _field_becomes_null_unique(self, old_field, new_field):
        return not (old_field.null and old_field.unique) and new_field.null and new_field.unique



from django.db.models import BooleanField, IntegerField, Lookup
from django.db.models.aggregates import Avg, Count, StdDev, Variance
from django.db.models.expressions import Value, OrderBy, OrderByList, Exists, RawSQL, Window, ExpressionList, Case, When, \
    DurationExpression, CombinedExpression
from django.db.models.fields.json import HasKeyLookup
from django.db.models.functions import Now, ATan2, Chr, Collate, Greatest, Least, Length, LPad, Random, \
    Repeat, RPad, StrIndex, Substr, Log, Ln, Mod, Round, Degrees, Power, Radians, RowNumber
from django.db.models.lookups import BuiltinLookup


def as_sqlserver(expression):
    def decorator(func):
        setattr(expression, 'as_sqlserver', func)
        return func
    return decorator


@as_sqlserver(Value)
def value_(self, compiler, connection, **extra):
    if compiler.query.group_by and self.value is not None and isinstance(self.output_field, IntegerField):
        # sql server would fail with group by col * %s so we replace literal
        return str(int(self.value)), ()
    return self.as_sql(compiler, connection, **extra)


@as_sqlserver(Now)
def now(self, compiler, connection, **extra):
    return self.as_sql(compiler, connection, template='SYSDATETIME()', **extra)


@as_sqlserver(Avg)
def avg(self, compiler, connection):
    return self.as_sql(compiler, connection, template='%(function)s(%(distinct)sCAST(%(field)s AS FLOAT))')


@as_sqlserver(Chr)
def chr(self, compiler, connection):
    return self.as_sql(compiler, connection, function='NCHAR')


@as_sqlserver(Collate)
def collate(self, compiler, connection, **extra_context):
    extra_context.setdefault('collation', self.collation)
    return self.as_sql(compiler, connection, **extra_context)


@as_sqlserver(Count)
def count(self, compiler, connection):
    return self.as_sql(compiler, connection, function='COUNT_BIG')


@as_sqlserver(Degrees)
@as_sqlserver(Radians)
def degrad(self, compiler, connection):
    float_tpl = '%(function)s(CAST(%(field)s AS FLOAT))'
    return self.as_sql(compiler, connection, template=float_tpl)


@as_sqlserver(Greatest)
def greatest(self, compiler, connection):
    template = '(SELECT MAX(value) FROM (VALUES (%(expressions)s)) AS _%(function)s(value))'
    return self.as_sql(compiler, connection, arg_joiner='), (', template=template)


@as_sqlserver(Least)
def least(self, compiler, connection):
    template = '(SELECT MIN(value) FROM (VALUES (%(expressions)s)) AS _%(function)s(value))'
    return self.as_sql(compiler, connection, arg_joiner='), (', template=template)


@as_sqlserver(Length)
def length_fn(self, compiler, connection):
    return self.as_sql(compiler, connection, function='LEN')


@as_sqlserver(LPad)
def lpad(self, compiler, connection):
    function = 'dbo.django_lpad'
    return self.as_sql(compiler, connection, function=function)


# @as_sqlserver(MD5)
# def md5(self, compiler, connection):
#     template = "LOWER(CONVERT(VARCHAR(32), HashBytes('%(function)s', CAST(%(expressions)s AS VARCHAR)), 2))"
#     return self.as_sql(compiler, connection, template=template)

#
# @as_sqlserver(SHA1)
# def sha1(self, compiler, connection):
#     template = "CONVERT(VARCHAR(32), HashBytes('%(function)s', %(expressions)s), 2)"
#     return self.as_sql(compiler, connection, template=template)


@as_sqlserver(Random)
def random(self, compiler, connection):
    return self.as_sql(compiler, connection, function='RAND')


@as_sqlserver(Repeat)
def repeat(self, compiler, connection):
    return self.as_sql(compiler, connection, function='REPLICATE')


@as_sqlserver(RPad)
def rpad(self, compiler, connection):
    function = 'dbo.django_rpad'
    return self.as_sql(compiler, connection, function=function)


@as_sqlserver(StdDev)
def stddev(self, compiler, connection):
    function = 'STDEV'
    if self.function == 'STDDEV_POP':
        function = 'STDEVP'
    return self.as_sql(compiler, connection, function=function)


@as_sqlserver(StrIndex)
def strindex(self, compiler, connection):
    clone = self.copy()
    clone.source_expressions.reverse()
    sql = clone.as_sql(compiler, connection, function='CHARINDEX')
    return sql


@as_sqlserver(Substr)
def substr(self, compiler, connection, **extra):
    if len(self.get_source_expressions()) < 3:
        clone = self.copy()
        clone.get_source_expressions().append(Value(2**31 - 1))
        return clone.as_sql(compiler, connection, **extra)
    return self.as_sql(compiler, connection, **extra)


@as_sqlserver(Variance)
def variance(self, compiler, connection):
    function = 'VAR'
    if self.function == 'VAR_POP':
        function = '%sP' % function
    return self.as_sql(compiler, connection, function=function)


@as_sqlserver(ATan2)
def atan2(self, compiler, connection, **extra_context):
    return self.as_sql(compiler, connection, function='ATN2', **extra_context)


@as_sqlserver(Log)
def log(self, compiler, connection, **extra_context):
    clone = self.copy()
    clone.set_source_expressions(self.get_source_expressions()[::-1])
    return clone.as_sql(compiler, connection, **extra_context)


@as_sqlserver(Ln)
def ln(self, compiler, connection, **extra_context):
    return self.as_sql(compiler, connection, function='LOG', **extra_context)


@as_sqlserver(Mod)
def sqlserver_mod(self, compiler, connection, **extra_context):
    compiler.escape_if_noparams = True
    return self.as_sql(compiler, connection, template='%(expressions)s', arg_joiner=' %% ', **extra_context)


@as_sqlserver(Power)
def sqlserver_pow(self, compiler, connection, **extra_context):
    template = '%(function)s(CAST(%(expressions)s)'
    return self.as_sql(compiler, connection, **extra_context, template=template, arg_joiner=' AS FLOAT),')


@as_sqlserver(Round)
def sqlserver_round(self, compiler, connection, **extra_context):
    return self.as_sql(compiler, connection, template='%(function)s(%(expressions)s, 0)', **extra_context)


@as_sqlserver(OrderBy)
def orderby(self, compiler, connection, **extra):
    if self.nulls_last or self.nulls_first:
        # sql server throws error "A constant expression was encountered in the ORDER BY" otherwise
        if isinstance(self.expression, RawSQL) and compiler._re_constant.match(self.expression.sql):
            return '%s %s' % (self.expression.sql, 'DESC' if self.descending else 'ASC'), ()
        isnull = 0 if self.nulls_first else 1
        isnotnull = 1 if self.nulls_first else 0
        template = 'IIF(%%(expression)s IS NULL, %d, %d) ASC, %%(expression)s %%(ordering)s' % (isnull, isnotnull)
        copy = self.copy()
        copy.nulls_last = copy.nulls_first = False  # otherwise as_sql overwrites template
        return copy.as_sql(compiler, connection, template=template, **extra)

    if isinstance(self.expression, (CombinedExpression, BuiltinLookup, Exists)) and \
            not isinstance(self.expression, (DurationExpression,)) and \
            isinstance(self.expression.output_field, BooleanField):
        # problem with DurationExpression is that output_field is calculated too early if we check it here
        copy = self.copy()
        copy.expression = Case(When(self.expression, then=True), default=False)
        return copy.as_sql(compiler, connection, **extra)
    return self.as_sql(compiler, connection, **extra)


@as_sqlserver(HasKeyLookup)
def has_key_lookup(self, compiler, connection):
    return self.as_sql(compiler, connection, template="JSON_VALUE(%s,  %%s) IS NOT NULL")


@as_sqlserver(Window)
def window(self, compiler, connection, **extra):
    if self.order_by is None and isinstance(self.source_expression, RowNumber):
        copy = self.copy()
        copy.order_by = OrderByList(*[expr for expr, _ in compiler.get_order_by()])
        return copy.as_sql(compiler, connection, **extra)
    return self.as_sql(compiler, connection, **extra)


@as_sqlserver(Lookup)
def lookup_fn(self, compiler, connection):
    # mostly copied from oracle
    compiler.escape_if_noparams = True
    wrapped = False
    exprs = []
    for expr in (self.lhs, self.rhs):
        if connection.ops.conditional_expression_supported_in_where_clause(expr) or isinstance(expr, (BuiltinLookup,)):
            expr = Case(When(expr, then=True), default=False)
            wrapped = True
        exprs.append(expr)
    lookup = type(self)(*exprs) if wrapped else self
    return lookup.as_sql(compiler, connection)



from django.db.backends.base.validation import BaseDatabaseValidation
from django.core import checks


class DatabaseValidation(BaseDatabaseValidation):

    def check_field_type(self, field, field_type):
        errors = []

        if field.db_index and field_type.upper() in self.connection._limited_data_types:
            errors.append(
                checks.Warning(
                    '%s does not support a database index on %s columns.'
                    % (self.connection.display_name, field_type),
                    hint=(
                        "An index won't be created. Silence this warning if "
                        "you don't care about it."
                    ),
                    obj=field,
                    id='fields.W162',
                )
            )
        return errors



from django.db.models.query import QuerySet
from django.db import connections


_bulk = QuerySet.bulk_update


def bulk_update(self, objs, fields, batch_size=None):
    affected = 0
    if connections[self.db].vendor == 'sqlserver':
        if batch_size is not None and batch_size < 0:
            raise ValueError('Batch size must be a positive integer.')
        if not fields:
            raise ValueError('Field names must be given to bulk_update().')
        objs = tuple(objs)
        if any(obj.pk is None for obj in objs):
            raise ValueError('All bulk_update() objects must have a primary key set.')
        field_names = {self.model._meta.get_field(name): name for name in fields}
        if any(not f.concrete or f.many_to_many for f in field_names):
            raise ValueError('bulk_update() can only be used with concrete fields.')
        if any(f.primary_key for f in field_names):
            raise ValueError('bulk_update() cannot be used with primary key fields.')
        if not objs:
            return 0
        for obj in objs:
            obj._prepare_related_fields_for_save(
                operation_name="bulk_update", fields=[self.model._meta.get_field(name) for name in fields]
            )
        same_values = {}
        for field in field_names:
            first = ok = True
            value = None
            for o in objs:
                attr = getattr(o, field.attname)
                if not first and value != attr:
                    ok = False
                    break
                if first:
                    first = False
                    value = attr
            if ok:
                same_values[field.attname] = value
        if len(same_values):
            affected = self.filter(pk__in={o.pk for o in objs}).update(**same_values)
            fields = [v for k, v in field_names.items() if k.attname not in same_values]

    if len(fields):
        return affected + _bulk(self, objs, fields, batch_size)
    return affected


bulk_update.alters_data = True

setattr(QuerySet, 'bulk_update', bulk_update)



from collections import namedtuple
from django.db.backends.base.introspection import BaseDatabaseIntrospection, TableInfo, FieldInfo as BaseFieldInfo

from django.db.models.indexes import Index

from tds_django.sql.queries import Introspection

FieldInfo = namedtuple('FieldInfo', BaseFieldInfo._fields + ('identity', 'seed', 'increment',))


class DatabaseIntrospection(BaseDatabaseIntrospection):
    # Map type codes to Django Field types.
    data_types_reverse = {48: 'SmallIntegerField', 52: 'SmallIntegerField', 56: 'IntegerField', 127: 'BigIntegerField',
                          50: 'BooleanField', 59: 'FloatField', 62: 'FloatField', 122: 'DecimalField',
                          60: 'DecimalField', 175: 'CharField', 167: 'CharField', 239: 'CharField', 231: 'CharField',
                          35: 'TextField', 99: 'TextField', 173: 'BinaryField', 165: 'BinaryField', 34: 'ImageField',
                          108: 'DecimalField', 106: 'DecimalField', 40: 'DateField', 41: 'TimeField',
                          42: 'DateTimeField', 43: 'DurationField', 58: 'DateTimeField', 61: 'DateTimeField',
                          36: 'UUIDField'}

    ignored_tables = []

    def get_field_type(self, data_type, description):
        field_type = super().get_field_type(data_type, description)
        if description.identity:
            if field_type == 'BigIntegerField':
                return 'BigAutoField'
            if field_type == 'SmallIntegerField':
                return 'SmallAutoField'
            return 'AutoField'
        return field_type

    def get_table_list(self, cursor):
        """
        Returns a list of table and view names in the current database.
        """
        cursor.execute(Introspection.table_list)
        types = {'BASE TABLE': 't', 'VIEW': 'v'}
        return [TableInfo(row[0], types.get(row[1]))
                for row in cursor.fetchall()
                if row[0] not in self.ignored_tables]

    def get_table_description(self, cursor, table_name):
        """ As of django 3.2 django still does not take defaults into account
            Ignores collation if it is the server default
        """
        cursor.execute(Introspection.table_description, (table_name, ))
        extras = {line[0]: line[1:] for line in cursor.fetchall()}

        cursor.execute('SELECT TOP 1 * FROM %s' % self.connection.ops.quote_name(table_name, ))
        return [FieldInfo(*column, *extras[column[0]]) for column in cursor.description]

    def get_sequences(self, cursor, table_name, table_fields=()):
        cursor.execute(Introspection.sequences, (table_name, ))
        # SQL Server allows only one identity column per table
        # https://docs.microsoft.com/en-us/sql/t-sql/statements/create-table-transact-sql-identity-property
        row = cursor.fetchone()
        return [{'table': table_name, 'column': row[0]}] if row else []

    def get_relations(self, cursor, table_name):
        cursor.execute(Introspection.relations, (table_name, ))
        return {item[0]: (item[2], item[1]) for item in cursor.fetchall()}

    def get_constraints(self, cursor, table_name):
        """
        Retrieves any constraints or keys (unique, pk, fk, check, index)
        across one or more columns.

        Returns a dict mapping constraint names to their attributes,
        where attributes is a dict with keys:
         * columns: List of columns this covers
         * primary_key: True if primary key, False otherwise
         * unique: True if this is a unique constraint, False otherwise
         * foreign_key: (table, column) of target, or None
         * check: True if check constraint, False otherwise
         * index: True if index, False otherwise.
         * orders: The order (ASC/DESC) defined for the columns of indexes
         * type: The type of the index (btree, hash, etc.)
        """
        constraints = {}
        # Loop over the key table, collecting things as constraints
        # This will get PKs, FKs, and uniques, but not CHECK
        cursor.execute(Introspection.get_constraints, (table_name, ))
        for constraint, column, kind, ref_table, ref_column in cursor.fetchall():
            # If we're the first column, make the record
            if constraint not in constraints:
                constraints[constraint] = {
                    'columns': [],
                    'primary_key': kind.lower() == 'primary key',
                    'unique': kind.lower() in ['primary key', 'unique'],
                    'foreign_key': (ref_table, ref_column) if kind.lower() == 'foreign key' else None,
                    'check': False,
                    'index': False,
                }
            # Record the details
            constraints[constraint]['columns'].append(column)
        # Now get CHECK constraint columns
        cursor.execute(Introspection.get_checks, (table_name, ))
        for constraint, column, clause in cursor.fetchall():
            # If we're the first column, make the record
            if constraint not in constraints:
                constraints[constraint] = {
                    'columns': [],
                    'primary_key': False,
                    'unique': False,
                    'foreign_key': None,
                    'check': True,
                    'index': False,
                    'clause': clause,
                }
            # Record the details
            constraints[constraint]['columns'].append(column)
        # indices
        cursor.execute(Introspection.get_indices, (table_name, ))
        indices = {}
        for name, unique, primary, type_, desc, order, has_filter, filter_definition, column in cursor.fetchall():
            if name not in indices:
                indices[name] = {
                    'columns': [],
                    'primary_key': primary,
                    'unique': unique,
                    'foreign_key': None,
                    'check': False,
                    'index': True,
                    'has_filter': has_filter,
                    'filter_definition': filter_definition,
                    'orders': [],
                    'type': Index.suffix if type_ in (1, 2) else desc.lower(),
                }
            indices[name]['columns'].append(column)
            indices[name]['orders'].append('DESC' if order == 1 else 'ASC')
        for name, constraint in indices.items():
            if name not in constraints:
                constraints[name] = constraint
        return constraints

    def get_table_defaults(self, cursor, table_name):
        cursor.execute(Introspection.get_default, (table_name, ))
        return {line[0]: line[1:] for line in cursor.fetchall()}



import re

from itertools import chain
from django.db.models.expressions import Subquery, RawSQL
from django.db.models.sql import compiler


class SQLCompiler(compiler.SQLCompiler):
    _re_constant = re.compile(r'^\s*\(?\s*\d+\s*\)?\s*')

    def as_sql(self, with_limits=True, with_col_aliases=False):
        sql, params = super().as_sql(with_limits=with_limits, with_col_aliases=with_col_aliases)
        if with_limits and not self.query.low_mark and self.query.high_mark is not None:
            if self.query.combinator == 'union':
                sql = f'SELECT TOP {self.query.high_mark} * FROM ({sql}) t'
            else:
                needle = 'SELECT DISTINCT' if self.query.distinct else 'SELECT'
                sql = sql.replace(needle, needle + ' TOP %d' % self.query.high_mark, 1)
        if not params and hasattr(self, 'escape_if_noparams'):
            sql = sql % ()
        return sql, params

    def get_order_by(self):
        order_by = super().get_order_by()
        if order_by:
            # TODO? sql server cannot order by constants
            pass
        elif self.query.low_mark:  # if offset, need order
            order_by = [(None, ('1 ASC', [], None)), ]
        return order_by

    def collapse_group_by(self, expressions, having):
        expressions = super().collapse_group_by(expressions, having)

        def allowed(e):
            if isinstance(e, RawSQL):
                return not self._re_constant.match(e.sql)
            if not (hasattr(e, 'contains_column_references') and e.contains_column_references):
                return False
            if isinstance(e, Subquery):
                return False
            return True

        return [e for e in expressions if allowed(e)]


class SQLInsertCompiler(compiler.SQLInsertCompiler, SQLCompiler):
    def fix_auto(self, sql, opts, fields, qn):
        if opts.auto_field is not None:
            # db_column is None if not explicitly specified by model field
            auto_field_column = opts.auto_field.db_column or opts.auto_field.column
            columns = [f.column for f in fields]
            if auto_field_column in columns:
                id_insert_sql = []
                table = qn(opts.db_table)
                sql_format = 'SET IDENTITY_INSERT %s ON; %s; SET IDENTITY_INSERT %s OFF'
                for q, p in sql:
                    id_insert_sql.append((sql_format % (table, q, table), p))
                sql = id_insert_sql

        return sql

    def as_sql(self):
        result = self._as_sql()
        if self.query.fields:
            # remove id if explicit insert default
            opts = self.query.get_meta()
            result = self.fix_auto(result, opts, self.query.fields or [opts.pk], self.connection.ops.quote_name)
        return result

    def _as_sql(self):
        # We don't need quote_name_unless_alias() here, since these are all
        # going to be column names (so we can avoid the extra overhead).
        qn = self.connection.ops.quote_name
        opts = self.query.get_meta()
        insert_statement = self.connection.ops.insert_statement(
            on_conflict=self.query.on_conflict,
        )
        result = ["%s %s" % (insert_statement, qn(opts.db_table))]

        # PATCH START for sql server we ignore identity fields differently
        if self.query.fields:
            fields = self.query.fields
            result.append('(%s)' % ', '.join(qn(f.column) for f in fields))
            value_rows = [
                [self.prepare_value(field, self.pre_save_val(field, obj)) for field in fields]
                for obj in self.query.objs
            ]
        else:
            # An empty object.
            value_rows = []
            fields = []
        # PATCH END


        # Currently the backends just accept values when generating bulk
        # queries and generate their own placeholders. Doing that isn't
        # necessary and it should be possible to use placeholders and
        # expressions in bulk inserts too.
        can_bulk = (
            not self.returning_fields and self.connection.features.has_bulk_insert
        )

        placeholder_rows, param_rows = self.assemble_as_sql(fields, value_rows)

        on_conflict_suffix_sql = self.connection.ops.on_conflict_suffix_sql(
            fields,
            self.query.on_conflict,
            self.query.update_fields,
            self.query.unique_fields,
        )
        if (
            self.returning_fields
            and self.connection.features.can_return_columns_from_insert
        ):
            if self.connection.features.can_return_rows_from_bulk_insert:
                result.append(
                    self.connection.ops.bulk_insert_sql(fields, placeholder_rows)
                )
                params = param_rows
            # PATCH START
            elif self.query.fields:
                result.append('VALUES (%s)' % ', '.join(placeholder_rows[0]))
                params = [param_rows[0]]
            else:
                result.append('DEFAULT VALUES')
                params = []
            # PATCH END

            if on_conflict_suffix_sql:
                result.append(on_conflict_suffix_sql)
            # Skip empty r_sql to allow subclasses to customize behavior for
            # 3rd party backends. Refs #19096.
            r_sql, self.returning_params = self.connection.ops.return_insert_columns(
                self.returning_fields
            )
            if r_sql:
                # PATCH START sql needs output before values
                result.insert(-1, r_sql)
                # PATCH END
                params += [self.returning_params]
            return [(" ".join(result), tuple(chain.from_iterable(params)))]

        if can_bulk:
            # PATCH START
            if not self.query.fields:
                result.append('DEFAULT VALUES')
                sql = ';'.join(' '.join(result) for _ in self.query.objs)
                return [(sql, ())]
            # PATCH END

            result.append(self.connection.ops.bulk_insert_sql(fields, placeholder_rows))
            if on_conflict_suffix_sql:
                result.append(on_conflict_suffix_sql)
            return [(" ".join(result), tuple(p for ps in param_rows for p in ps))]
        else:
            if on_conflict_suffix_sql:
                result.append(on_conflict_suffix_sql)
            return [
                (" ".join(result + ["VALUES (%s)" % ", ".join(p)]), vals)
                for p, vals in zip(placeholder_rows, param_rows)
            ]


class SQLDeleteCompiler(compiler.SQLDeleteCompiler, SQLCompiler):
    def as_sql(self):
        sql, params = super().as_sql()
        if sql:
            sql = '; '.join(['SET NOCOUNT OFF', sql])
        return sql, params


class SQLUpdateCompiler(compiler.SQLUpdateCompiler, SQLCompiler):
    def as_sql(self):
        sql, params = super().as_sql()
        if sql:
            sql = '; '.join(['SET NOCOUNT OFF', sql])
        return sql, params


class SQLAggregateCompiler(compiler.SQLAggregateCompiler, SQLCompiler):
    pass



import pytds
from django.core.exceptions import ImproperlyConfigured
from django.conf import settings
from django.db import IntegrityError
from django.db.backends.base.base import BaseDatabaseWrapper
from django.utils.asyncio import async_unsafe
from .client import DatabaseClient
from .creation import DatabaseCreation
from .features import DatabaseFeatures
from .introspection import DatabaseIntrospection
from .operations import DatabaseOperations
from .schema import DatabaseSchemaEditor
from .validation import DatabaseValidation


class DatabaseWrapper(BaseDatabaseWrapper):
    display_name = 'tds-django'
    vendor = 'sqlserver'
    Database = pytds

    data_types = {
        'AutoField': 'INT',
        'BigAutoField': 'BIGINT',
        'BigIntegerField': 'BIGINT',
        'BinaryField': 'VARBINARY(MAX)',
        'BooleanField': 'BIT',
        'CharField': 'NVARCHAR(%(max_length)s)',
        'DateField': 'DATE',
        'DateTimeField': 'DATETIME2',
        'DecimalField': 'DECIMAL(%(max_digits)s, %(decimal_places)s)',
        'DurationField': 'BIGINT',
        'FileField': 'NVARCHAR(%(max_length)s)',
        'FilePathField': 'NVARCHAR(%(max_length)s)',
        'FloatField': 'DOUBLE PRECISION',
        'IntegerField': 'INT',
        'IPAddressField': 'NVARCHAR(15)',
        'GenericIPAddressField': 'NVARCHAR(39)',
        'JSONField': 'NVARCHAR(MAX)',
        'NullBooleanField': 'BIT',
        'OneToOneField': 'INT',
        'PositiveBigIntegerField': 'BIGINT',
        'PositiveIntegerField': 'INT',
        'PositiveSmallIntegerField': 'SMALLINT',
        'SlugField': 'NVARCHAR(%(max_length)s)',
        'SmallAutoField': 'SMALLINT',
        'SmallIntegerField': 'SMALLINT',
        'TextField': 'NVARCHAR(MAX)',
        'TimeField': 'TIME',
        'UUIDField': 'UNIQUEIDENTIFIER',
    }

    _limited_data_types = ('NVARCHAR(MAX)',)

    data_types_suffix = {
        'AutoField': 'IDENTITY (1, 1)',
        'BigAutoField': 'IDENTITY (1, 1)',
        'SmallAutoField': 'IDENTITY (1, 1)',
    }

    data_type_check_constraints = {
        'PositiveIntegerField': '[%(column)s] >= 0',
        'PositiveSmallIntegerField': '[%(column)s] >= 0',
    }
    operators = {
        # Since '=' is used not only for string comparison there is no way
        # to make it case (in)sensitive.
        'exact': '= %s',
        'iexact': "= UPPER(%s)",
        'contains': "LIKE %s ESCAPE '\\'",
        'icontains': "LIKE UPPER(%s) ESCAPE '\\'",
        'gt': '> %s',
        'gte': '>= %s',
        'lt': '< %s',
        'lte': '<= %s',
        'startswith': "LIKE %s ESCAPE '\\'",
        'endswith': "LIKE %s ESCAPE '\\'",
        'istartswith': "LIKE UPPER(%s) ESCAPE '\\'",
        'iendswith': "LIKE UPPER(%s) ESCAPE '\\'",
    }

    # copied from pg
    pattern_esc = r"REPLACE(REPLACE(REPLACE({}, '\', '[\]'), '%%', '[%%]'), '_', '[_]')"
    pattern_ops = {
        'contains': "LIKE '%%' + {} + '%%'",
        'icontains': "LIKE '%%' + UPPER({}) + '%%'",
        'startswith': "LIKE {} + '%%'",
        'istartswith': "LIKE UPPER({}) + '%%'",
        'endswith': "LIKE '%%' + {}",
        'iendswith': "LIKE '%%' + UPPER({})",
    }

    SchemaEditorClass = DatabaseSchemaEditor

    client_class = DatabaseClient
    creation_class = DatabaseCreation
    features_class = DatabaseFeatures
    introspection_class = DatabaseIntrospection
    ops_class = DatabaseOperations
    validation_class = DatabaseValidation

    def get_connection_params(self):
        settings_dict = self.settings_dict
        # TODO warnings for user
        conn_params = {
            'dsn': settings_dict['HOST'] or 'localhost',
            'port': settings_dict['PORT'] or 1433,
            'database': settings_dict['NAME'],
            'user': settings_dict['USER'],
            'password': settings_dict['PASSWORD'],
            'autocommit': getattr(settings, 'AUTOCOMMIT', False),
            'use_mars': False,
         }
        return conn_params

    def get_new_connection(self, conn_params):
        conn = self.Database.connect(**conn_params)
        return conn

    def init_connection_state(self):
        pass

    def _set_autocommit(self, autocommit):
        with self.wrap_database_errors:
            self.connection.autocommit = autocommit

    @async_unsafe
    def create_cursor(self, name=None):
        return self.connection.cursor()

    def _savepoint_commit(self, sid):
        pass

    def disable_constraint_checking(self):
        with self.cursor() as cursor:
            cursor.execute('EXEC sp_MSforeachtable "ALTER TABLE ? NOCHECK CONSTRAINT ALL";')
        return True

    def enable_constraint_checking(self):
        self.needs_rollback, needs_rollback = False, self.needs_rollback
        try:
            with self.cursor() as cursor:
                # !! we do not check when re-enabling constraint !!
                cursor.execute('EXEC sp_MSforeachtable "ALTER TABLE ? WITH NOCHECK CHECK CONSTRAINT ALL";')
        finally:
            self.needs_rollback = needs_rollback

    def check_constraints(self, table_names=None):
        with self.cursor() as cursor:
            self.clear_connection_messages()
            if table_names is None:
                sql = ['DBCC CHECKCONSTRAINTS WITH ALL_CONSTRAINTS, NO_INFOMSGS']
            else:
                sql = ["DBCC CHECKCONSTRAINTS ('%s') WITH ALL_CONSTRAINTS, NO_INFOMSGS" % t for t in table_names]
            for s in sql:
                cursor.execute(s)
                if cursor.description:
                    r = cursor.fetchone()
                    raise IntegrityError(r)

    def is_usable(self):
        return not self.connection._closed

    def get_connection_messages(self, cursor):
        return cursor.messages

    def clear_connection_messages(self):
        pass



from django.db.backends.base.features import BaseDatabaseFeatures
from django.utils.functional import cached_property


class DatabaseFeatures(BaseDatabaseFeatures):

    supports_timezones = False  # has datetimeoffset but would break things if mixed with datetimefield2?

    # json
    supports_json_field = False
    can_introspect_json_field = False  # ISJSON only validates json objects, not literals
    has_json_object_function = False
    supports_json_field_contains = False
    
    allow_sliced_subqueries_with_in = False  # TODO CHECK
    can_create_inline_fk = False

    can_return_columns_from_insert = True
    can_return_id_from_insert = True
    can_rollback_ddl = True
    # can_use_chunked_reads = False
    greatest_least_ignores_nulls = True
    has_bulk_insert = True

    has_native_uuid_field = True
    has_real_datatype = True
    # has_zoneinfo_database = False
    # ignores_table_name_case = False  # depends on db collation
    implied_column_null = True
    max_query_params = 1000
    requires_literal_defaults = True

    # supported_explain_formats = {'XML', }
    supports_boolean_expr_in_select_clause = False
    supports_covering_indexes = True
    supports_expression_indexes = False  # would need to manage computed column
    supports_ignore_conflicts = False
    supports_index_on_text_field = False
    supports_over_clause = True
    supports_paramstyle_pyformat = False  # TODO
    supports_partially_nullable_unique_constraints = False
    supports_sequence_reset = False  # TODO
    supports_subqueries_in_group_by = False
    supports_temporal_subtraction = True
    supports_transactions = True

    ignores_unnecessary_order_by_in_subqueries = False

    test_db_allows_multiple_connections = False

    django_test_skips = {
        "no need to test this": {
          'cache.tests.FileBasedCacheTests',
          'cache.tests.FileBasedCachePathLibTests',
          'cache.tests.LocMemCacheTests',
        },
        "test assumption not correct for SQL Server": {
            # sql server allows double quoted table names
            'schema.tests.SchemaTests.test_add_foreign_key_quoted_db_table',
            # hardcoded SUBSTR
            'custom_lookups.tests.BilateralTransformTests.test_transform_order_by',
            # hardcoded RETURNING
            'queries.test_db_returning.ReturningValuesTests.test_insert_returning',
            'queries.test_db_returning.ReturningValuesTests.test_insert_returning_multiple',
            # expects LIMIT but sql server has TOP
            'queries.test_qs_combinators.QuerySetSetOperationTests.test_exists_union',
        },
        "tests for test db creation. Fail because we override create_test_db": {
            'backends.base.test_creation.TestDbCreationTests.test_mark_expected_failures_and_skips_call',
            'backends.base.test_creation.TestDbCreationTests.test_migrate_test_setting_false',
            'backends.base.test_creation.TestDbCreationTests.test_migrate_test_setting_true',
        },
        "SQL Server need more queries than what is expected (deleting and recreating indexes)": {
            'schema.tests.SchemaTests.test_unique_and_reverse_m2m',
            'schema.tests.SchemaTests.test_unique_no_unnecessary_fk_drops',
            'admin_views.tests.GroupAdminTest.test_group_permission_performance',
            'admin_views.tests.UserAdminTest.test_user_permission_performance',
        },
        "test uses cursor directly and we don't fix the query, the query should set (SET IDENTITY_INSERT off)": {
            'migrations.test_operations.OperationTests.test_create_model_with_constraint',
            'migrations.test_operations.OperationTests.test_create_model_with_boolean_expression_in_check_constraint'
        },
        "unsupported by SQL Server": {
            # subquery in agg function
            'aggregation.tests.AggregateTestCase.test_aggregation_subquery_annotation_values_collision',
            'aggregation.test_filter_argument.FilteredAggregateTests.test_filtered_aggregate_on_exists',
            'db_functions.math.test_mod.ModTests.test_float',
            # subquery in avg
            'expressions_case.tests.CaseExpressionTests.test_annotate_with_in_clause',
            # ORDER BY LOB in window function
            'expressions_window.tests.WindowFunctionTests.test_key_transform',
            # NTH_RESULT: could be done
            'expressions_window.tests.WindowFunctionTests.test_nth_returns_null',
            'expressions_window.tests.WindowFunctionTests.test_nthvalue',
            # order by constant
            'ordering.tests.OrderingTests.test_order_by_constant_value',
            # distinct + subquery, could work if rewrite distinct with group by
            'ordering.tests.OrderingTests.test_orders_nulls_first_on_filtered_subquery',
            # sql server does not allow like predicates index
            'indexes.tests.PartialIndexTests.test_multiple_conditions',
        },
        "Test is using hardcoded values that are different for sql server": {
            'aggregation.tests.AggregateTestCase.test_count_star',
            'cache.tests.CreateCacheTableForDBCacheTests.test_createcachetable_observes_database_router',
        },
        "Avg are cast as float, with can cause issues with decimals": {
            'aggregation_regress.tests.AggregationTests.test_values_list_annotation_args_ordering',
        },
        "client unsupported": {
            "dbshell.tests.DbshellCommandTestCase.test_command_missing",
        },
        "I don't understand": {

            'model_fields.test_decimalfield.DecimalFieldTests.test_roundtrip_with_trailing_zeros',
            # also: extra "is an old API that we aim to deprecate"
            'queries.tests.Queries5Tests.test_extra_select_literal_percent_s',

            # problems with unicode? hash functions in sql server return different result from python
            'db_functions.text.test_md5.MD5Tests.test_basic',
            'db_functions.text.test_md5.MD5Tests.test_transform',
            'db_functions.text.test_sha1.SHA1Tests.test_basic',
            'db_functions.text.test_sha1.SHA1Tests.test_transform',
            'db_functions.text.test_sha224.SHA224Tests.test_basic',
            'db_functions.text.test_sha224.SHA224Tests.test_transform',
            'db_functions.text.test_sha256.SHA256Tests.test_basic',
            'db_functions.text.test_sha256.SHA256Tests.test_transform',
            'db_functions.text.test_sha384.SHA384Tests.test_basic',
            'db_functions.text.test_sha384.SHA384Tests.test_transform',
            'db_functions.text.test_sha512.SHA512Tests.test_basic',
            'db_functions.text.test_sha512.SHA512Tests.test_transform',
            # problem with iterator: django/db/models/query.py list(islice(iterator, chunk_size))
            # only returs first chunk
            'prefetch_related.tests.PrefetchRelatedTests.test_m2m_prefetching_iterator_with_chunks',
            #
            'backends.base.test_creation.TestDbCreationTests.test_migrate_test_setting_true',
            'backends.base.test_creation.TestDbCreationTests.test_migrate_test_setting_false_ensure_schema',
        },
        "To check?": {
            # update pk field means updating all FK: doable but I will never do that automatically
            'migrations.test_operations.OperationTests.test_alter_field_pk_fk_db_collation',
            'schema.tests.SchemaTests.test_alter_primary_key_db_collation',
            # pass with wrong error without mars, not pass on mars, pytds bug?
            'backends.tests.BackendTestCase.test_cursor_contextmanager',
            # TODO check more
            'expressions.tests.FTimeDeltaTests.test_durationfield_multiply_divide',
            # doable but low priority
            'queries.test_bulk_update.BulkUpdateTests.test_updated_rows_when_passing_duplicates',
            # TODO json, currenlty marked as not available in features
            # 'model_fields.test_jsonfield.TestQuerying.test_has_any_keys',
            # 'model_fields.test_jsonfield.TestQuerying.test_array_key_contains',
            # 'model_fields.test_jsonfield.TestQuerying.test_contained_by',
            # 'model_fields.test_jsonfield.TestQuerying.test_contains_contained_by_with_key_transform',
            # 'model_fields.test_jsonfield.TestQuerying.test_contains_primitives',
            # 'model_fields.test_jsonfield.TestQuerying.test_deep_lookup_array',
            # 'model_fields.test_jsonfield.TestQuerying.test_deep_lookup_mixed',
            # 'model_fields.test_jsonfield.TestQuerying.test_deep_lookup_objs',
            # 'model_fields.test_jsonfield.TestQuerying.test_deep_lookup_transform',
            # 'model_fields.test_jsonfield.TestQuerying.test_exact',
            # 'model_fields.test_jsonfield.TestQuerying.test_exact_complex',
            # 'model_fields.test_jsonfield.TestQuerying.test_expression_wrapper_key_transform',
            # 'model_fields.test_jsonfield.TestQuerying.test_join_key_transform_annotation_expression',
            # 'model_fields.test_jsonfield.TestQuerying.test_isnull_key_or_none',
            # 'model_fields.test_jsonfield.TestQuerying.test_isnull',
            # 'model_fields.test_jsonfield.TestQuerying.test_deep_values',
            # 'model_fields.test_jsonfield.TestQuerying.test_contains',
            # 'model_fields.test_jsonfield.TestQuerying.test_has_key',
            # 'model_fields.test_jsonfield.TestQuerying.test_has_key_deep',
            # 'model_fields.test_jsonfield.TestQuerying.test_has_key_list',
            # 'model_fields.test_jsonfield.TestQuerying.test_has_key_null_value',
            # 'model_fields.test_jsonfield.TestQuerying.test_has_keys',
            # 'model_fields.test_jsonfield.TestQuerying.test_key_iregex',
            # 'model_fields.test_jsonfield.TestQuerying.test_key_quoted_string',
            # 'model_fields.test_jsonfield.TestQuerying.test_key_regex',
            # 'model_fields.test_jsonfield.TestQuerying.test_lookups_with_key_transform',
            # 'model_fields.test_jsonfield.TestQuerying.test_order_grouping_custom_decoder',
            # 'model_fields.test_jsonfield.TestQuerying.test_ordering_grouping_by_count',
            # 'model_fields.test_jsonfield.TestQuerying.test_ordering_grouping_by_key_transform',
            # 'model_fields.test_jsonfield.JSONFieldTests.test_db_check_constraints',
            # 'model_fields.test_jsonfield.TestQuerying.test_isnull_key',
            # 'model_fields.test_jsonfield.TestQuerying.test_key_in',
            # 'model_fields.test_jsonfield.TestQuerying.test_key_transform_expression',
            # 'model_fields.test_jsonfield.TestQuerying.test_key_values',
            # 'model_fields.test_jsonfield.TestQuerying.test_nested_key_transform_expression',
            # 'model_fields.test_jsonfield.TestQuerying.test_none_key',
            # 'model_fields.test_jsonfield.TestQuerying.test_none_key_and_exact_lookup',
            # 'model_fields.test_jsonfield.TestQuerying.test_none_key_exclude',
            # 'model_fields.test_jsonfield.TestQuerying.test_ordering_by_transform',
            # 'model_fields.test_jsonfield.TestQuerying.test_shallow_lookup_obj_target',
        },
        "SQL Server does not natively support unique (with multiple) nullable fields so a FK to such field will fail": {
            "many_to_one_null.tests.ManyToOneNullTests.test_add_efficiency",
            "many_to_one_null.tests.ManyToOneNullTests.test_assign_clear_related_set",
            "many_to_one_null.tests.ManyToOneNullTests.test_assign_with_queryset",
            "many_to_one_null.tests.ManyToOneNullTests.test_clear_efficiency",
            "many_to_one_null.tests.ManyToOneNullTests.test_created_via_related_set",
            "many_to_one_null.tests.ManyToOneNullTests.test_created_without_related",
            "many_to_one_null.tests.ManyToOneNullTests.test_get_related",
            "many_to_one_null.tests.ManyToOneNullTests.test_related_null_to_field",
            "many_to_one_null.tests.ManyToOneNullTests.test_related_set",
            "many_to_one_null.tests.ManyToOneNullTests.test_remove_from_wrong_set",
            "many_to_one_null.tests.ManyToOneNullTests.test_set",
            "many_to_one_null.tests.ManyToOneNullTests.test_set_clear_non_bulk",
        }
    }

    # Collation names for use by the Django test suite.
    test_collations = {
        'ci': 'Latin1_General_CI_AS',  # Case-insensitive.
        'cs': 'Latin1_General_BIN2',  # Case-sensitive.
        'non_default': 'Korean_90_BIN',  # Non-default.
        'swedish_ci': 'Finnish_Swedish_CI_AI'  # Swedish case-insensitive.
    }

    @cached_property
    def introspected_field_types(self):
        return {
            **super().introspected_field_types,
            'GenericIPAddressField': 'CharField',
            'PositiveBigIntegerField': 'BigIntegerField',
            'PositiveIntegerField': 'IntegerField',
            'PositiveSmallIntegerField': 'SmallIntegerField',
            'DurationField': 'BigIntegerField',
        }



import re
from django.conf import settings
from django.db import DatabaseError, OperationalError
from django.db.backends.base.operations import BaseDatabaseOperations
from django.db.models import Exists, ExpressionWrapper, Lookup
from django.db.models.expressions import RawSQL
from django.db.models.sql.where import WhereNode
from django.utils import timezone

from tds_django.sql.queries import Introspection
from .tz import iana_win_map


class DatabaseOperations(BaseDatabaseOperations):
    cast_char_field_without_max_length = 'NVARCHAR(MAX)'
    compiler_module = 'tds_django.compiler'
    # explain_prefix = 'SET SHOWPLAN_XML ON;'
    _re_utc_offset = re.compile(r'^utc[+-]', re.IGNORECASE)

    def savepoint_create_sql(self, sid):
        return 'SAVE TRANSACTION %s' % self.quote_name(sid)

    def savepoint_rollback_sql(self, sid):
        return 'ROLLBACK TRANSACTION %s' % self.quote_name(sid)

    def _prepare_tzname_delta(self, tzname):
        if '/' in tzname:
            t = iana_win_map.get(tzname, False)
            if not t:
                raise ValueError(f"Invalid TimeZone {tzname}")
            return t
        elif self._re_utc_offset.match(tzname):
            return tzname[3:]
        return tzname

    def _sql_tz(self, sql, params, tzname):
        tzname = self._prepare_tzname_delta(tzname)
        if tzname[0] in '+-':
            return f"SWITCHOFFSET({sql}, '{tzname}')", params
        else:
            return f"{sql} AT TIME ZONE '{tzname}'", params

    def _convert_sql_to_tz(self, sql, params, tzname):
        if tzname and settings.USE_TZ and self.connection.timezone_name != tzname:
            sql, params = self._sql_tz(sql, params, self.connection.timezone_name or 'UTC')
            return self._sql_tz(sql, params, tzname)
        return sql, params

    def datetime_extract_sql(self, lookup_type, sql, params, tzname):
        sql, params = self._convert_sql_to_tz(sql, params, tzname)
        return self.date_extract_sql(lookup_type, sql, params)


    def date_extract_sql(self, lookup_type, sql, params):
        # TODO GET SCHEMA. IN CONNECTION?
        if not lookup_type.lower() in 'year quarter month dayofyear day weekday hour minute week second ' \
                'millisecond microsecond nanosecond tzoffset iso_week iso_week_day iso_year week_day'.split():
            raise OperationalError(f'Lookup {lookup_type} not supported.')
        return f"dbo.django_date_extract('{lookup_type.lower()}', {sql})", params

    def datetime_cast_date_sql(self, sql, params, tzname):
        sql, params = self._convert_sql_to_tz(sql, params, tzname)
        return f'CAST({sql} AS DATE)', params

    def date_trunc_sql(self, lookup_type, sql, params, tzname=None):
        sql, params = self.datetime_trunc_sql(lookup_type, sql, params, tzname)
        return f'CAST({sql} AS DATE)', params

    def datetime_trunc_sql(self, lookup_type, sql, params, tzname=None):
        sql, params = self._convert_sql_to_tz(sql, params, tzname)
        if not lookup_type.lower() in 'year quarter month dayofyear day weekday hour minute week second'.split():
            raise OperationalError(f'Lookup {lookup_type} not supported.')
        return f"dbo.django_datetime_trunc('{lookup_type}', {sql})", params

    def datetime_cast_time_sql(self, sql, params, tzname):
        sql, params = self._convert_sql_to_tz(sql, params, tzname)
        return f'CAST({sql} as TIME)', params

    def time_trunc_sql(self, lookup_type, sql, params, tzname=None):
        """ similar mysql """
        sql, params = self._convert_sql_to_tz(sql, params, tzname)
        fields = {
            'hour': r'hh\:\0\0\:\0\0',
            'minute': r'hh\:mm\:\0\0',
            'second': r'hh\:mm\:ss',
        }
        if lookup_type in fields:
            return f"CAST(FORMAT(CAST({sql} AS TIME), '{fields[lookup_type]}') AS TIME)", params
        else:
            return f'CAST({sql} as TIME)', params

    def quote_name(self, name):
        """
        Returns a quoted version of the given table, index or column name. Does
        not quote the given name if it's already been quoted.
        """
        if name.startswith('[') and name.endswith(']'):
            return name
        return '[%s]' % name

    def bulk_insert_sql(self, fields, placeholder_rows):
        placeholder_rows_sql = (", ".join(row) for row in placeholder_rows)
        values_sql = ", ".join("(%s)" % sql for sql in placeholder_rows_sql)
        return "VALUES " + values_sql

    def sql_flush(self, style, tables, *, reset_sequences=False, allow_cascade=False):
        if not tables:
            return []
        sql = ['EXEC sp_MSforeachtable "ALTER TABLE ? NOCHECK CONSTRAINT ALL"']

        if allow_cascade:
            with self.connection.cursor() as cursor:
                cursor.execute(Introspection.get_fks)
                m = {(a, b): (c, d) for a, b, c, d in cursor.fetchall()}
                cols_to_del = set()
                while True:
                    newm = {}
                    for (s, k), (t, c) in m.items():
                        if t in tables or (t, c) in cols_to_del:
                            sql.append('DELETE FROM %s WHERE %s IS NOT NULL' % (self.quote_name(s),
                                                                                  self.quote_name(k)))
                            cols_to_del.add((s, k))
                        else:
                            newm[(s, k)] = (t, c)
                    if len(newm) == len(m):
                        break
                    m = newm

        sql.extend(
            '%s %s %s;' % (
                style.SQL_KEYWORD('DELETE'),
                style.SQL_KEYWORD('FROM'),
                style.SQL_FIELD(self.quote_name(table_name)),
            ) for table_name in tables
        )
        if reset_sequences:
            qids = Introspection.get_identities % \
                   ', '.join('%s' for _ in tables)
            with self.connection.cursor() as cursor:
                cursor.execute(qids, tuple(tables))
                for t, i in cursor.fetchall():
                    sql.append('DBCC CHECKIDENT (%s, RESEED, %d)' % (self.quote_name(t), i))

        sql += ['EXEC sp_MSforeachtable "ALTER TABLE ? WITH CHECK CHECK CONSTRAINT ALL";']
        return sql

    def get_db_converters(self, expression):
        converters = super().get_db_converters(expression)
        internal_type = expression.output_field.get_internal_type()
        if internal_type == 'DateTimeField':
            converters.append(self.convert_datetimefield_value)
        elif internal_type == 'FloatField':
            converters.append(self.convert_floatfield_value)
        elif internal_type in ('BooleanField', 'NullBooleanField'):
            converters.append(self.convert_booleanfield_value)
        return converters

    def convert_datetimefield_value(self, value, expression, connection):
        if value is not None:
            if settings.USE_TZ and not timezone.is_aware(value):
                value = timezone.make_aware(value, self.connection.timezone)
        return value

    def convert_floatfield_value(self, value, expression, connection):
        if value is not None:
            value = float(value)
        return value

    def convert_booleanfield_value(self, value, expression, connection):
        return bool(value) if value in (0, 1) else value

    def conditional_expression_supported_in_where_clause(self, expression):
        """ same as oracle """
        if isinstance(expression, (Exists, Lookup, WhereNode,)):
            return True
        if isinstance(expression, ExpressionWrapper) and expression.conditional:
            return self.conditional_expression_supported_in_where_clause(expression.expression)
        if isinstance(expression, RawSQL) and expression.conditional:
            return True
        return False

    def max_name_length(self):
        return 128

    def return_insert_columns(self, fields):
        if not fields:
            return '', ()
        field_names = []
        params = []
        for field in fields:
            field_names.append('INSERTED.%s' % self.quote_name(field.column))
            params.append(field)
        return 'OUTPUT %s' % ', '.join(field_names), ()

    def fetch_returned_insert_columns(self, cursor, returning_params):
        return cursor.fetchone()

    def pk_default_value(self):
        """
        Return the value to use during an INSERT statement to specify that
        the field should use its default value.
        """
        return 'NULL'

    def check_expression_support(self, expression):
        """It feels useless to send less info to the user than what would be returned from sql server. """
        pass

    def _get_limit_offset_params(self, low_mark, high_mark):
        offset = low_mark or 0
        if high_mark is not None:
            return (high_mark - offset), offset
        return None, offset

    def limit_offset_sql(self, low_mark, high_mark):
        limit, offset = self._get_limit_offset_params(low_mark, high_mark)
        if limit and not offset:
            return ''  # need TOP x, done in compiler
        return ' '.join(sql for sql in (
            ('OFFSET %d ROWS' % offset) if offset else None,
            ('FETCH NEXT %d ROWS ONLY' % limit) if limit else None,
        ) if sql)

    def combine_expression(self, connector, sub_expressions):
        """
        SQL Server requires special cases for some operators in query expressions
        """
        if connector == '^':
            return 'POWER(CAST(%s AS FLOAT), %s)' % tuple(sub_expressions)
        elif connector == '<<':
            return 'dbo.django_bitshift(%s, -1 * %s, 0)' % tuple(sub_expressions)
        elif connector == '>>':
            return 'dbo.django_bitshift(%s, %s, 0)' % tuple(sub_expressions)
        elif connector == '#':
            return super().combine_expression('^', sub_expressions)
        return super().combine_expression(connector, sub_expressions)

    def subtract_temporals(self, internal_type, lhs, rhs):
        lhs_sql, lhs_params = lhs
        rhs_sql, rhs_params = rhs
        return f'DATEDIFF_BIG(microsecond, {rhs_sql}, {lhs_sql})', (*rhs_params, *lhs_params)

    def combine_duration_expression(self, connector, sub_expressions):
        if connector not in ['+', '-',]:
            raise DatabaseError('Invalid connector for timedelta: %s.' % connector)
        lhs, rhs = sub_expressions
        sign = '-' if connector == '-' else ''
        if lhs.startswith('dbo'):
            col, sql = rhs, lhs
        else:
            col, sql = lhs, rhs
        return sql.format(col, sign)

    def format_for_duration_arithmetic(self, sql):
        return 'dbo.django_dtdelta({}, {}%s)' % sql

    def adapt_timefield_value(self, value):
        if value is None:
            return None
        # Expression values are adapted by the database.
        if hasattr(value, 'resolve_expression'):
            return value
        if timezone.is_aware(value):
            raise ValueError("SQL Server backend does not support timezone-aware times.")
        return str(value)

    def adapt_datetimefield_value(self, value):
        """
        Transforms a datetime value to an object compatible with what is expected
        by the backend driver for datetime columns.
        """
        if value is None:
            return None

        if hasattr(value, 'resolve_expression'):
            return value

        if timezone.is_aware(value):
            if settings.USE_TZ:
                value = value.astimezone(self.connection.timezone).replace(tzinfo=None)
            else:
                raise ValueError("SQL Server backend does not support timezone-aware datetimes when USE_TZ is False.")

        return value

    def bulk_batch_size(self, fields, objs):
        if fields:
            return self.connection.features.max_query_params // len(fields)
        return len(objs)

    def cache_key_culling_sql(self):
        return 'SELECT [cache_key] FROM %s ORDER BY cache_key OFFSET %%s ROWS FETCH FIRST 1 ROWS ONLY'

    def last_executed_query(self, cursor, sql, params):
        if params:
            m = tuple(f'{self.connection.SchemaEditorClass.quote_value(p) if p is not None else "NULL"}' for p in params)
            return sql % m
        return sql

    def lookup_cast(self, lookup_type, internal_type=None):
        if lookup_type in ('iexact', 'icontains', 'istartswith', 'iendswith'):
            return 'UPPER(%s)'
        if internal_type == 'UUIDField':
            return 'LOWER(%s)'
        return '%s'

    def prep_for_iexact_query(self, x):
        return x

    def regex_lookup(self, lookup_type):
        if lookup_type == 'regex':
            return 'dbo.django_regex(%s, %s) = 1'
        return 'dbo.django_iregex(%s, %s) = 1'



from setuptools import setup, find_packages

with open("README.md", "r", encoding="utf-8") as fh:
    long_description = fh.read()

setup(
    name="tds-django",
    version="4.1.0",
    author="Etienne Cogels",
    author_email="ecogels@users.noreply.github.com",
    description="Django backend for SQL Server using tds",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/ecogels/tds-django",
    project_urls={
        "Bug Tracker": "https://github.com/ecogels/tds-django/issues",
    },
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: BSD License",
        "Operating System :: OS Independent",
        "Development Status :: 4 - Beta",
        "Framework :: Django",
        "Framework :: Django :: 3.2",
        "Framework :: Django :: 4.0",
        "Programming Language :: Python",
        "Topic :: Database",
        "Topic :: Software Development :: Libraries",
    ],
    packages=find_packages(),
    include_package_data=True,
    package_data={'': ['sql/*.sql']},
    python_requires=">=3.8",
)




class Introspection:
    table_list = 'SELECT TABLE_NAME, TABLE_TYPE FROM INFORMATION_SCHEMA.TABLES WHERE TABLE_SCHEMA = SCHEMA_NAME()'

    table_description = """
SELECT c.name, OBJECT_DEFINITION(c.default_object_id),
     IIF(c.collation_name = CAST(DATABASEPROPERTYEX(DB_NAME(),'collation') AS VARCHAR(256)), NULL, c.collation_name),
     c.is_identity, i.seed_value, i.increment_value
FROM sys.objects o
     INNER JOIN sys.columns c ON o.object_id = c.object_id
     LEFT JOIN sys.identity_columns i ON c.object_id = i.object_id
WHERE o.schema_id = SCHEMA_ID() AND o.name = %s"""

    sequences = """
SELECT c.name FROM sys.columns c
INNER JOIN sys.tables t ON c.object_id = t.object_id
WHERE t.schema_id = SCHEMA_ID() AND t.name = %s AND c.is_identity = 1"""

    relations = """
SELECT e.COLUMN_NAME AS column_name , c.TABLE_NAME AS referenced_table_name, d.COLUMN_NAME AS referenced_column_name
FROM INFORMATION_SCHEMA.TABLE_CONSTRAINTS AS a
INNER JOIN INFORMATION_SCHEMA.REFERENTIAL_CONSTRAINTS AS b
    ON a.CONSTRAINT_NAME = b.CONSTRAINT_NAME AND a.TABLE_SCHEMA = b.CONSTRAINT_SCHEMA
INNER JOIN INFORMATION_SCHEMA.CONSTRAINT_TABLE_USAGE AS c
    ON b.UNIQUE_CONSTRAINT_NAME = c.CONSTRAINT_NAME AND b.CONSTRAINT_SCHEMA = c.CONSTRAINT_SCHEMA
INNER JOIN INFORMATION_SCHEMA.CONSTRAINT_COLUMN_USAGE AS d
    ON c.CONSTRAINT_NAME = d.CONSTRAINT_NAME AND c.CONSTRAINT_SCHEMA = d.CONSTRAINT_SCHEMA
INNER JOIN INFORMATION_SCHEMA.CONSTRAINT_COLUMN_USAGE AS e
    ON a.CONSTRAINT_NAME = e.CONSTRAINT_NAME AND a.TABLE_SCHEMA = e.TABLE_SCHEMA
WHERE a.TABLE_SCHEMA = SCHEMA_NAME() AND a.TABLE_NAME = %s AND a.CONSTRAINT_TYPE = 'FOREIGN KEY'"""

    key_columns = """
SELECT c.name AS column_name, rt.name AS referenced_table_name, rc.name AS referenced_column_name
FROM sys.foreign_key_columns fk
INNER JOIN sys.tables t ON t.object_id = fk.parent_object_id
INNER JOIN sys.columns c ON c.object_id = t.object_id AND c.column_id = fk.parent_column_id
INNER JOIN sys.tables rt ON rt.object_id = fk.referenced_object_id
INNER JOIN sys.columns rc ON rc.object_id = rt.object_id AND rc.column_id = fk.referenced_column_id
WHERE t.schema_id = SCHEMA_ID() AND t.name = %s"""

    get_constraints = """
SELECT
    kc.constraint_name,
    kc.column_name,
    tc.constraint_type,
    fk.referenced_table_name,
    fk.referenced_column_name
FROM INFORMATION_SCHEMA.KEY_COLUMN_USAGE AS kc
INNER JOIN INFORMATION_SCHEMA.TABLE_CONSTRAINTS AS tc ON
    kc.table_schema = tc.table_schema AND
    kc.table_name = tc.table_name AND
    kc.constraint_name = tc.constraint_name
LEFT OUTER JOIN (
    SELECT
        ps.name AS table_schema,
        pt.name AS table_name,
        pc.name AS column_name,
        rt.name AS referenced_table_name,
        rc.name AS referenced_column_name
    FROM
        sys.foreign_key_columns fkc
    INNER JOIN sys.tables pt ON fkc.parent_object_id = pt.object_id
    INNER JOIN sys.schemas ps ON pt.schema_id = ps.schema_id
    INNER JOIN sys.columns pc ON fkc.parent_object_id = pc.object_id AND
        fkc.parent_column_id = pc.column_id
    INNER JOIN sys.tables rt ON fkc.referenced_object_id = rt.object_id
    INNER JOIN sys.schemas rs ON rt.schema_id = rs.schema_id
    INNER JOIN sys.columns rc ON
        fkc.referenced_object_id = rc.object_id AND
        fkc.referenced_column_id = rc.column_id
) fk ON
    kc.table_schema = fk.table_schema AND
    kc.table_name = fk.table_name AND
    kc.column_name = fk.column_name
WHERE
    kc.table_schema = SCHEMA_NAME() AND
    kc.table_name = %s
ORDER BY
    kc.constraint_name ASC,
    kc.ordinal_position ASC """

    get_checks = """
SELECT kc.constraint_name, kc.column_name, cc.CHECK_CLAUSE
FROM INFORMATION_SCHEMA.CONSTRAINT_COLUMN_USAGE AS kc
    JOIN INFORMATION_SCHEMA.TABLE_CONSTRAINTS AS c ON
        kc.table_schema = c.table_schema AND
        kc.table_name = c.table_name AND
        kc.constraint_name = c.constraint_name
    JOIN INFORMATION_SCHEMA.CHECK_CONSTRAINTS cc ON
        cc.CONSTRAINT_CATALOG = c.CONSTRAINT_CATALOG
        AND cc.CONSTRAINT_NAME = c.CONSTRAINT_NAME
WHERE kc.table_schema = SCHEMA_NAME()
    AND kc.table_name =  %s """

    get_indices = """
SELECT
    i.name AS index_name,
    i.is_unique,
    i.is_primary_key,
    i.type,
    i.type_desc,
    ic.is_descending_key,
    i.has_filter,
    i.filter_definition,
    c.name AS column_name
FROM
    sys.tables AS t
INNER JOIN sys.schemas AS s ON
    t.schema_id = s.schema_id
INNER JOIN sys.indexes AS i ON
    t.object_id = i.object_id
INNER JOIN sys.index_columns AS ic ON
    i.object_id = ic.object_id AND
    i.index_id = ic.index_id
INNER JOIN sys.columns AS c ON
    ic.object_id = c.object_id AND
    ic.column_id = c.column_id
WHERE
    t.schema_id = SCHEMA_ID() AND
    t.name = %s
ORDER BY
    i.index_id ASC,
    ic.index_column_id ASC """

    get_default = """
SELECT c.name, d.name, d.definition
FROM sys.tables t
JOIN sys.default_constraints d ON d.parent_object_id = t.object_id
JOIN sys.columns c ON c.object_id = t.object_id AND c.column_id = d.parent_column_id
WHERE t.name = %s
AND t.schema_id = SCHEMA_ID()"""

    get_fks = """
SELECT OBJECT_NAME(f.parent_object_id) source_table,
    COL_NAME(fc.parent_object_id, fc.parent_column_id) source_col,
    OBJECT_NAME(f.referenced_object_id) target_table,
    COL_NAME(fc.referenced_object_id, fc.referenced_column_id) target_col
FROM sys.foreign_keys AS f
    INNER JOIN sys.foreign_key_columns AS fc ON f.object_id = fc.constraint_object_id
WHERE f.schema_id = SCHEMA_ID()"""

    get_identities = """
SELECT o.name, i.seed_value FROM sys.objects o
INNER JOIN sys.columns c ON o.object_id = c.object_id
LEFT JOIN sys.identity_columns i ON c.object_id = i.object_id
WHERE o.schema_id = SCHEMA_ID() AND c.is_identity = 1 AND o.name IN (%s)"""


class Misc:
    delete_table = """
DECLARE @query NVARCHAR(MAX) = N'';
SELECT @query += N' ALTER TABLE ' + QUOTENAME(cs.name) + '.' + QUOTENAME(ct.name)  + ' DROP CONSTRAINT ' +
QUOTENAME(fk.name) + ';'  FROM sys.foreign_keys AS fk 
INNER JOIN sys.tables AS ct ON fk.parent_object_id = ct.[object_id]
INNER JOIN sys.schemas AS cs ON ct.[schema_id] = cs.[schema_id]
WHERE fk.referenced_object_id = OBJECT_ID('%(table)s');
EXEC(@query);
DROP TABLE %(table)s """
